"""Client for connecting to SETRA API"""
import json
import logging
import os
import urllib.parse
from typing import Tuple, Union, List

import requests

from setra_client.models import Multi

logger = logging.getLogger(__name__)


def load_json_file(name):
    """Load json file from the fixtures directory"""
    here = os.path.realpath(
        os.path.join(os.getcwd(),
                     os.path.dirname(__file__).rsplit('/', 1)[0]))
    with open(os.path.join(here, 'tests/fixtures', name)) as f:
        data = json.load(f)
    return data


def merge_dicts(*dicts):
    """
    Combine a series of dicts without mutating any of them.

    >>> merge_dicts({'a': 1}, {'b': 2})
    {'a': 1, 'b': 2}
    >>> merge_dicts({'a': 1}, {'a': 2})
    {'a': 2}
    >>> merge_dicts(None, None, None)
    {}
    """
    combined = dict()
    for d in dicts:
        if not d:
            continue
        for k in d:
            combined[k] = d[k]
    return combined


class SetraEndpoints:
    def __init__(self,
                 url,
                 batch_url='batch/',
                 transaction_url='transaction/',
                 voucher_url='voucher/',
                 multi_url='addtrans/'
                 ):
        self.baseurl = url
        self.batch_url = batch_url
        self.transaction_url = transaction_url
        self.voucher_url = voucher_url
        self.multi_url = multi_url

    """ Get endpoints relative to the SETRA API URL. """

    def __repr__(self):
        return '{cls.__name__}({url!r})'.format(
            cls=type(self),
            url=self.baseurl)

    def batch(self, batch_id: str = None):
        """
        URL for Batch endpoint
        """
        if batch_id is None:
            return urllib.parse.urljoin(self.baseurl, self.batch_url)
        else:
            return urllib.parse.urljoin(self.baseurl,
                                        '/'.join((self.batch_url, batch_id)))

    def transaction(self, trans_id: str = None):
        """
        Url for Transaction endpoint
        """
        if trans_id is None:
            return urllib.parse.urljoin(self.baseurl, self.transaction_url)
        else:
            return urllib.parse.urljoin(self.baseurl,
                                        '/'.join((self.transaction_url,
                                                  trans_id)))

    def voucher(self, vouch_id: str = None):
        """
        Url for Voucher endpoint
        """
        if vouch_id is None:
            return urllib.parse.urljoin(self.baseurl, self.voucher_url)
        else:
            return urllib.parse.urljoin(self.baseurl,
                                        '/'.join((self.voucher_url, vouch_id)))

    def post_multi(self):
        return urllib.parse.urljoin(self.baseurl, self.multi_url)


class SetraClient(object):
    default_headers = {
        'Accept': 'application/json',
    }

    def __init__(self,
                 url: str,
                 headers: Union[None, dict] = None,
                 return_objects: bool = True,
                 use_sessions: bool = True,
                 ):
        """
        SETRA API client.

        :param str url: Base API URL
        :param dict headers: Append extra headers to all requests
        :param bool return_objects: Return objects instead of raw JSON
        :param bool use_sessions: Keep HTTP connections alive (default True)
        """

        self.urls = SetraEndpoints(url)
        self.headers = merge_dicts(self.default_headers, headers)
        self.return_objects = return_objects
        if use_sessions:
            self.session = requests.Session()
        else:
            self.session = requests

    def _build_request_headers(self, headers):
        request_headers = {}
        for h in self.headers:
            request_headers[h] = self.headers[h]
        for h in (headers or ()):
            request_headers[h] = headers[h]
        return request_headers

    def call(self,
             method_name,
             url,
             headers=None,
             params=None,
             return_response=True,
             **kwargs):
        headers = self._build_request_headers(headers)
        if params is None:
            params = {}
        logger.debug('Calling %s %s with params=%r',
                     method_name,
                     urllib.parse.urlparse(url).path,
                     params)
        r = self.session.request(method_name,
                                 url,
                                 headers=headers,
                                 params=params,
                                 **kwargs)
        if r.status_code in (500, 400, 401):
            logger.warning('Got HTTP %d: %r', r.status_code, r.content)
        if return_response:
            return r
        r.raise_for_status()
        return r.json()

    def get(self, url, **kwargs):
        return self.call('GET', url, **kwargs)

    def put(self, url, **kwargs):
        return self.call('PUT', url, **kwargs)

    def post(self, url, **kwargs):
        return self.call('POST', url, **kwargs)

    def object_or_data(self, cls, data) -> Union[object, dict]:
        if not self.return_objects:
            return data
        return cls.from_dict(data)

    def get_batch(self, batch_id: int = None):
        """
        GETs one or all batches from SETRA
        """
        url = self.urls.batch(str(batch_id))
        response = self.get(url)
        return response.json()

    def get_voucher(self, vouch_id: int):
        """
        GETs one or all batches from SETRA
        """
        url = self.urls.voucher(str(vouch_id))
        response = self.get(url)
        return response.json()

    def get_transaction(self, trans_id: int):
        """
        GETs one or all batches from SETRA
        """
        url = self.urls.transaction(str(trans_id))
        response = self.get(url)
        return response.json()

    def post_multi(self, multidata: Multi):
        """
        POST combination of batch, vouchers and transactions
        """
        url = self.urls.post_multi()
        headers = {'Content-Type': 'application/json'}
        response = self.post(url,
                             data=multidata.json(),
                             headers=headers)
        return response


def get_client(config_dict):
    """
    Get a SetraClient from configuration.
    """
    return SetraClient(**config_dict)
